"""Routines for handling etrees representing VOEvent packets."""
from __future__ import absolute_import
from __future__ import unicode_literals
from six import string_types
from lxml import objectify, etree
import collections

import voeventparse.definitions

voevent_v2_0_schema = etree.XMLSchema(
    etree.fromstring(voeventparse.definitions.v2_0_schema_str))


def Voevent(stream, stream_id, role):
    """Create a new VOEvent element tree, with specified IVORN and role.

    Args:

        stream (string): used to construct the IVORN like so::

                ivorn = 'ivo://' + stream + '#' + stream_id

            (N.B. ``stream_id`` is converted to string if required.)
            So, e.g. we might set::

                stream='voevent.soton.ac.uk/super_exciting_events'
                stream_id=77

        stream_id (string): See above.

        role (string): role as defined in VOEvent spec.
            (See also  :py:class:`.definitions.roles`)

    Returns:
        Root-node of the VOEvent, as represented by an lxml.objectify element
            tree ('etree'). See also
            http://lxml.de/objectify.html#the-lxml-objectify-api
    """
    parser = objectify.makeparser(remove_blank_text=True)
    v = objectify.fromstring(voeventparse.definitions.v2_0_skeleton_str,
                             parser=parser)
    _remove_root_tag_prefix(v)
    if not isinstance(stream_id, string_types):
        stream_id = repr(stream_id)
    v.attrib['ivorn'] = ''.join(('ivo://', stream, '#', stream_id))
    v.attrib['role'] = role
    # Presumably we'll always want the following children:
    # (NB, valid to then leave them empty)
    etree.SubElement(v, 'Who')
    etree.SubElement(v, 'What')
    etree.SubElement(v, 'WhereWhen')
    v.Who.Description = ('VOEvent created with voevent-parse: '
                         + 'https://github.com/timstaley/voevent-parse')
    return v


def loads(s, check_version=True):
    """
    Load VOEvent from bytes.

    This parses a VOEvent XML packet string, taking care of some subtleties.
    For Python 3 users, ``s`` should be a bytes object - see also
    http://lxml.de/FAQ.html,
    "Why can't lxml parse my XML from unicode strings?"
    (Python 2 users can stick with old-school ``str`` type if preferred)

    By default, will raise an exception if the VOEvent is not of version
    2.0. This can be disabled but voevent-parse routines are untested with
    other versions.

    Args:
        s (bytes): Bytes containing raw XML.
        check_version (bool): (Default=True) Checks that the VOEvent is of a
            supported schema version - currently only v2.0 is supported.
    Returns:
        voevent (:py:class:`Voevent`): Root-node of the  etree.
    Raises:
        exceptions.ValueError

    """
    # .. note::
    #
    # The namespace is removed from the root element tag to make
    #        objectify access work as expected,
    #        (see  :py:func:`._remove_root_tag_prefix`)
    #        so we must re-insert it when we want to conform to schema.
    v = objectify.fromstring(s)
    _remove_root_tag_prefix(v)

    if check_version:
        version = v.attrib['version']
        if not version == '2.0':
            raise ValueError('Unsupported VOEvent schema version:'+ version)

    return v


def load(file, check_version=True):
    """Load VOEvent from file object.

    A simple wrapper to read a file before passing the contents to
    :py:func:`.loads`. Use with an open file object, e.g.::

        with open('/path/to/voevent.xml', 'rb') as f:
            v = vp.load(f)

    Args:
        file (file): An open file object (binary mode preferred), see also
            http://lxml.de/FAQ.html :
            "Can lxml parse from file objects opened in unicode/text mode?"

        check_version (bool): (Default=True) Checks that the VOEvent is of a
            supported schema version - currently only v2.0 is supported.
    Returns:
        voevent (:py:class:`Voevent`): Root-node of the  etree.
    """
    s = file.read()
    return loads(s, check_version)


def dumps(voevent, pretty_print=False, xml_declaration=True, encoding='UTF-8'):
    """Converts voevent to string.

    .. note:: Default encoding is UTF-8, in line with VOE2.0 schema.
        Declaring the encoding can cause diffs with the original loaded VOEvent,
        but I think it's probably the right thing to do (and lxml doesn't
        really give you a choice anyway).

    Args:
        voevent (:class:`Voevent`): Root node of the VOevent etree.
        pretty_print (bool): indent the output for improved human-legibility
            when possible. See also:
            http://lxml.de/FAQ.html#why-doesn-t-the-pretty-print-option-reformat-my-xml-output
        xml_declaration (bool): Prepends a doctype tag to the string output,
            i.e. something like ``<?xml version='1.0' encoding='UTF-8'?>``
    Returns:
        Bytes containing raw XML representation of VOEvent.

    """
    _return_to_standard_xml(voevent)
    s = etree.tostring(voevent, pretty_print=pretty_print,
                       xml_declaration=xml_declaration,
                       encoding=encoding)
    _remove_root_tag_prefix(voevent)
    return s


def dump(voevent, file, pretty_print=True, xml_declaration=True):
    """Writes the voevent to the file object.

    e.g.::

        with open('/tmp/myvoevent.xml','wb') as f:
            voeventparse.dump(v, f)

    Args:
        voevent(:class:`Voevent`): Root node of the VOevent etree.
        file (file): An open (binary mode) file object for writing.
        pretty_print
        pretty_print(bool): See :func:`dumps`
        xml_declaration(bool): See :func:`dumps`
    """
    file.write(dumps(voevent, pretty_print, xml_declaration))


def valid_as_v2_0(voevent):
    """Tests if a voevent conforms to the schema.

    Args:
        voevent(:class:`Voevent`): Root node of a VOEvent etree.
    Returns: Bool (VOEvent is valid?)
    """
    _return_to_standard_xml(voevent)
    valid_bool = voevent_v2_0_schema.validate(voevent)
    _remove_root_tag_prefix(voevent)
    return valid_bool


def assert_valid_as_v2_0(voevent):
    """
    Raises :py:obj:`lxml.etree.DocumentInvalid` if voevent is invalid.

    Especially useful for debugging,
    since the stack trace contains a reason for the invalidation.

    Args:
        voevent(:class:`Voevent`): Root node of a VOEvent etree.
    Returns: None. NB raises :py:obj:`lxml.etree.DocumentInvalid` if VOEvent
        does not conform to schema.
    """
    _return_to_standard_xml(voevent)
    voevent_v2_0_schema.assertValid(voevent)
    _remove_root_tag_prefix(voevent)


def set_who(voevent, date=None, author_ivorn=None):
    """Sets the minimal 'Who' attributes:  date of authoring, AuthorIVORN.

    Args:
        voevent(:class:`Voevent`): Root node of a VOEvent etree.
        date(datetime.datetime): Date of authoring.
            NB Microseconds are ignored, as per the VOEvent spec.
        author_ivorn(string): Short author identifier,
            e.g. ``voevent.4pisky.org/ALARRM``.
            Note that the prefix ``ivo://`` will be prepended internally.

    """
    if author_ivorn is not None:
        voevent.Who.AuthorIVORN = ''.join(('ivo://', author_ivorn))
    if date is not None:
        voevent.Who.Date = date.replace(microsecond=0).isoformat()


def set_author(voevent, title=None, shortName=None, logoURL=None,
               contactName=None, contactEmail=None, contactPhone=None,
               contributor=None):
    """For setting fields in the detailed author description.

    This can optionally be neglected if a well defined AuthorIVORN is supplied.

    .. note:: Unusually for this library,
        the args here use CamelCase naming convention,
        since there's a direct mapping to the ``Author.*``
        attributes to which they will be assigned.

    Args:
        voevent(:class:`Voevent`): Root node of a VOEvent etree.
            The rest of the arguments are strings corresponding to child elements.
    """
    # We inspect all local variables except the voevent packet,
    # Cycling through and assigning them on the Who.Author element.
    AuthChildren = locals()
    AuthChildren.pop('voevent')
    if not voevent.xpath('Who/Author'):
        etree.SubElement(voevent.Who, 'Author')
    for k, v in AuthChildren.items():
        if v is not None:
            voevent.Who.Author[k] = v


def add_where_when(voevent, coords, obs_time, observatory_location):
    """Add details of an observation to the WhereWhen section.

    Args:
        voevent(:class:`Voevent`): Root node of a VOEvent etree.
        coords(:class:`.Position2D`): Sky co-ordinates of event.
        obs_time(datetime.datetime): Nominal DateTime of the observation.
        observatory_location(string): Telescope locale, e.g. 'La Palma'.
            May be a generic location as listed under
            :class:`voeventparse.definitions.observatory_location`.
    """

    # .. todo:: Implement TimeError using datetime.timedelta

    obs_data = etree.SubElement(voevent.WhereWhen, 'ObsDataLocation')
    etree.SubElement(obs_data, 'ObservatoryLocation', id=observatory_location)
    ol = etree.SubElement(obs_data, 'ObservationLocation')
    etree.SubElement(ol, 'AstroCoordSystem', id=coords.system)
    ac = etree.SubElement(ol, 'AstroCoords',
                          coord_system_id=coords.system)
    time = etree.SubElement(ac, 'Time', unit='s')
    instant = etree.SubElement(time, 'TimeInstant')
    instant.ISOTime = obs_time.isoformat()
    # iso_time = etree.SubElement(instant, 'ISOTime') = obs_time.isoformat()

    pos2d = etree.SubElement(ac, 'Position2D', unit=coords.units)
    pos2d.Name1 = 'RA'
    pos2d.Name2 = 'Dec'
    pos2d_val = etree.SubElement(pos2d, 'Value2')
    pos2d_val.C1 = coords.ra
    pos2d_val.C2 = coords.dec
    pos2d.Error2Radius = coords.err


def add_how(voevent, descriptions=None, references=None):
    """Add descriptions or references to the How section.

    Args:
        voevent(:class:`Voevent`): Root node of a VOEvent etree.
        descriptions(string): Description string, or list of description
            strings.
        references(:py:class:`voeventparse.misc.Reference`): A reference element
            (or list thereof).
    """
    if not voevent.xpath('How'):
        etree.SubElement(voevent, 'How')
    if descriptions is not None:
        for desc in _listify(descriptions):
            # d = etree.SubElement(voevent.How, 'Description')
            # voevent.How.Description[voevent.How.index(d)] = desc
            ##Simpler:
            etree.SubElement(voevent.How, 'Description')
            voevent.How.Description[-1] = desc
    if references is not None:
        voevent.How.extend(_listify(references))


def add_why(voevent, importance=None, expires=None, inferences=None):
    """Add Inferences, or set importance / expires attributes of the Why section.

    .. note::

        ``importance`` / ``expires`` are 'Why' attributes, therefore setting them
        will overwrite previous values.
        ``inferences``, on the other hand,  are appended to the list.

    Args:
        voevent(:class:`Voevent`): Root node of a VOEvent etree.
        importance(float): Value from 0.0 to 1.0
        expires(datetime.datetime): Expiration date given inferred reason
            (See voevent spec).
        inferences(:class:`voeventparse.misc.Inference`): Inference or list of
            inferences, denoting probable identifications or associations, etc.
    """
    if not voevent.xpath('Why'):
        etree.SubElement(voevent, 'Why')
    if importance is not None:
        voevent.Why.attrib['importance'] = str(importance)
    if expires is not None:
        voevent.Why.attrib['expires'] = expires.replace(
            microsecond=0).isoformat()
    if inferences is not None:
        voevent.Why.extend(_listify(inferences))


def add_citations(voevent, citations):
    """Add citations to other voevents.

    The schema mandates that the 'Citations' section must either be entirely
    absent, or non-empty - hence we require this wrapper function for its
    creation prior to listing the first citation.

    Args:
        voevent(:class:`Voevent`): Root node of a VOEvent etree.
        citation(:class:`voeventparse.misc.Citation`): Citation or list of citations.

    """
    if not voevent.xpath('Citations'):
        etree.SubElement(voevent, 'Citations')
    voevent.Citations.extend(_listify(citations))



# ###################################################
# And finally, utility functions...

def _remove_root_tag_prefix(v):
    """
    Removes 'voe' namespace prefix from root tag.

    When we load in a VOEvent, the root element has a tag prefixed by
     the VOE namespace, e.g. {http://www.ivoa.net/xml/VOEvent/v2.0}VOEvent
    Because objectify expects child elements to have the same namespace as
    their parent, this breaks the python-attribute style access mechanism.
    We can get around it without altering root, via e.g
     who = v['{}Who']

    Alternatively, we can temporarily ditch the namespace altogether.
    This makes access to elements easier, but requires care to reinsert
    the namespace upon output.

    I've gone for the latter option.
    """
    if v.prefix:
        # Create subelement without a prefix via etree.SubElement
        etree.SubElement(v, 'original_prefix')
        # Now carefully access said named subelement (without prefix cascade)
        # and alter the first value in the list of children with this name...
        # LXML syntax is a minefield!
        v['{}original_prefix'][0] = v.prefix
        v.tag = v.tag.replace(''.join(('{', v.nsmap[v.prefix], '}')), '')
        # Now v.tag = '{}VOEvent', v.prefix = None
    return


def _reinsert_root_tag_prefix(v):
    """
    Returns namespace prefix to root tag, if it had one.
    """
    if hasattr(v, 'original_prefix'):
        original_prefix = v.original_prefix
        del v.original_prefix
        v.tag = ''.join(('{', v.nsmap[original_prefix], '}VOEvent'))
    return


def _return_to_standard_xml(v):
    # Remove lxml.objectify DataType namespace prefixes:
    objectify.deannotate(v)
    #Put the default namespace back:
    _reinsert_root_tag_prefix(v)
    etree.cleanup_namespaces(v)


# Define this for convenience in add_how:
def _listify(x):
    """Ensure x is iterable; if not then enclose it in a list and return it."""
    if isinstance(x, string_types):
        return [x]
    elif isinstance(x, collections.Iterable):
        return x
    else:
        return [x]


"""Routines for creating sub-elements of the VOEvent tree,
and a few other helper classes."""

from __future__ import absolute_import
from six import string_types
from collections import namedtuple
import datetime
from lxml import objectify, etree


class Position2D(namedtuple('Position2D', 'ra dec err units system')):
    """A namedtuple for simple representation of a 2D position as described
    by the VOEvent spec.

    Args:
        ra (float): Right ascension.
        dec (float): Declination
        err (float): Error radius.
        units (:class:`.definitions.coord_units`): Coordinate units
            e.g. degrees, radians.
        system (:class:`.definitions.sky_coord_system`): Co-ordinate system
            e.g. UTC-FK5-GEO
    """
    pass  # Just wrapping a namedtuple so we can assign a docstring.


_datatypes_autoconversion = {
    bool: ('string', lambda b: str(b)),
    int: ('int', lambda i: str(i)),
    float: ('float', lambda f: str(f)),
    datetime.datetime: ('string', lambda dt: dt.isoformat()),
}


def Param(name, value=None, unit=None, ucd=None, dataType=None, utype=None,
          ac=True):
    """
    'Parameter', used as a general purpose key-value entry in the 'What' section.

    May be assembled into a :class:`Group`.

    NB ``name`` is not mandated by schema, but *is* mandated in full spec.

    Args:
        value(string): String representing parameter value.
            Or, if ``ac`` is true, then 'autoconversion' is attempted, in which case
            ``value`` can also be an instance of one of the following:

             * :py:obj:`bool`
             * :py:obj:`int`
             * :py:obj:`float`
             * :py:class:`datetime.datetime`

            This allows you to create Params without littering your code
            with string casts, or worrying if the passed value is a float or a
            string, etc.
            NB the value is always *stored* as a string representation,
            as per VO spec.
        unit(string): e.g. 'deg' for degrees.
        ucd(string): `unified content descriptor <http://arxiv.org/abs/1110.0525>`_.
            For a list of valid UCDs, see:
            http://vocabularies.referata.com/wiki/Category:IVOA_UCD.
        dataType(string): Denotes type of ``value``; restricted to 3 options:
            ``string`` (default), ``int`` , or ``float``.
            (NB *not* to be confused with standard XML Datatypes, which have many
            more possible values.)
        utype(string): See http://wiki.ivoa.net/twiki/bin/view/IVOA/Utypes
        ac(bool): Attempt automatic conversion of passed ``value`` to string,
            and set ``dataType`` accordingly (only attempted if ``dataType``
            is the default, i.e. ``None``).
            (NB only supports types listed in _datatypes_autoconversion dict)

    """
    # We use locals() to allow concise looping over the arguments.
    atts = locals()
    atts.pop('ac')
    temp_dict={}
    temp_dict.update(atts)
    for k in temp_dict.keys():
        if atts[k] is None:
            del atts[k]
    if (ac
        and value is not None
        and (not isinstance(value, string_types))
        and dataType is None
    ):
        if type(value) in _datatypes_autoconversion:
            datatype, func = _datatypes_autoconversion[type(value)]
            atts['dataType'] = datatype
            atts['value'] = func(value)
    return objectify.Element('Param', attrib=atts)


def Group(params, name=None, type=None):
    """Groups together Params for adding under the 'What' section.

    Args:
        params(list of :func:`Param`): Parameter elements to go in this group.
        name(string): Group name. NB ``None`` is valid, since the group may be
            best identified by its type.
        type(string): Type of group, e.g. 'complex' (for real and imaginary).
    """
    atts = {}
    if name:
        atts['name'] = name
    if type:
        atts['type'] = type
    g = objectify.Element('Group', attrib=atts)
    for p in params:
        g.append(p)
    return g


def Reference(uri, meaning=None):
    """
    Represents external information, typically original obs data and metadata.

    Args:
        uri(string): Uniform resource identifier for external data, e.g. FITS file.
        meaning(string): The nature of the document referenced, e.g. what
            instrument and filter was used to create the data?
    """
    attrib = {'uri': uri}
    if meaning is not None:
        attrib['meaning'] = meaning
    return objectify.Element('Reference', attrib)


def Inference(probability=None, relation=None, name=None, concept=None):
    """Represents a probable cause / relation between this event and some prior.

    Args:
        probability(float): Value 0.0 to 1.0.
        relation(string): e.g. 'associated' or 'identified' (see Voevent spec)
        name(string): e.g. name of identified progenitor.
        concept(string): One of a 'formal UCD-like vocabulary of astronomical
            concepts', e.g. http://ivoat.ivoa.net/stars.supernova.Ia - see
            VOEvent spec.
    """
    atts = {}
    if probability is not None:
        atts['probability'] = str(probability)
    if relation is not None:
        atts['relation'] = relation
    inf = objectify.Element('Inference', attrib=atts)
    if name is not None:
        inf.Name = name
    if concept is not None:
        inf.Concept = concept
    return inf


def Citation(ivorn, cite_type):
    """Used to cite earlier VOEvents.

    Args:
        ivorn(string): It is assumed this will be copied verbatim from elsewhere,
            and so these should have any prefix (e.g. 'ivo://','http://')
            already in place - the function will not alter the value.
        cite_type (:class:`.definitions.cite_types`): String conforming to one
            of the standard citation types.

    """
    # This is an ugly hack around the limitations of the  lxml.objectify API:
    c = objectify.StringElement(cite=cite_type)
    c._setText(ivorn)
    c.tag = "EventIVORN"
    return c




import unittest
from unittest import TestCase
from voeventparse.tests.resources import datapaths
from six import b
from lxml import objectify, etree
from copy import copy
import datetime

import voeventparse as vp


class TestValidation(TestCase):
    def shortDescription(self):
        return None

    def test_schema_valid_for_test_data(self):
        """
        First, let's check everything is in order with the test data
        and the schema.

        Here we directly access the lxml validate routine.
        """
        v = objectify.parse(datapaths.swift_bat_grb_pos_v2).getroot()
        self.assertTrue(vp.voevent_v2_0_schema.validate(v))
        Who = v['{}Who']
        Who.BadChild = 42
        self.assertFalse(vp.voevent_v2_0_schema.validate(v))
        del Who.BadChild
        self.assertTrue(vp.voevent_v2_0_schema.validate(v))

        # NB dropping the namespace from root element invalidates packet:
        # This is why we must re-insert it before output.
        v.tag = 'VOEvent'
        self.assertFalse(vp.voevent_v2_0_schema.validate(v))

    def test_validation_routine(self):
        """
        Now we perform the same validation tests, but applied via the
        convenience functions.
        """
        with open(datapaths.swift_bat_grb_pos_v2, 'rb') as f:
            v = vp.load(f)
        self.assertTrue(vp.valid_as_v2_0(v))
        v.Who.BadChild = 42
        self.assertFalse(vp.valid_as_v2_0(v))
        del v.Who.BadChild
        self.assertTrue(vp.valid_as_v2_0(v))

    def test_invalid_error_reporting(self):
        with self.assertRaises(etree.DocumentInvalid):
            v = vp.Voevent(stream='voevent.soton.ac.uk/TEST',
                            stream_id='001',
                            role='DeadParrot')
            vp.assert_valid_as_v2_0(v)


class TestIO(TestCase):
    def shortDescription(self):
        return None

    def test_load_of_voe_v2(self):
        with open(datapaths.swift_bat_grb_pos_v2, 'rb') as f:
            vff = vp.load(f)
        with open(datapaths.swift_bat_grb_pos_v2, 'rb') as f:
            vfs = vp.loads(f.read())
        self.assertEqual(objectify.dump(vff), objectify.dump(vfs))
        self.assertEqual(vfs.tag, 'VOEvent')
        self.assertEqual(vfs.attrib['ivorn'],
                         'ivo://nasa.gsfc.gcn/SWIFT#BAT_GRB_Pos_532871-729')

    def test_load_of_voe_v1(self):
        with self.assertRaises(ValueError):
            with open(datapaths.swift_xrt_pos_v1, 'rb') as f:
                vff = vp.load(f)

        #Can override version checking, at own risk!
        with open(datapaths.swift_xrt_pos_v1, 'rb') as f:
            vff = vp.load(f, check_version=False)

    def test_namespace_variations(self):
        # NB, not enclosing root element in a namespace is invalid under schema
        # But this has been seen in the past (isolated bug case?)
        # Anyway, handled easily enough
        with open(datapaths.no_namespace_test_packet, 'rb') as f:
            vff = vp.load(f)
        self.assertFalse(vp.valid_as_v2_0(vff))
        self.assertEqual(vff.tag, 'VOEvent')
        self.assertEqual(vff.attrib['ivorn'],
                         'ivo://com.dc3/dc3.broker#BrokerTest-2014-02-24T15:55:27.72')

        with open(datapaths.swift_bat_grb_pos_v2, 'rb') as f:
            xml_str = f.read()
            xml_str = xml_str.replace(b'voe', b'foobar_ns')
        # print xml_str
        vfs = vp.loads(xml_str)
        vp.assert_valid_as_v2_0(vfs)
        self.assertEqual(vfs.tag, 'VOEvent')
        self.assertEqual(vfs.attrib['ivorn'],
                         'ivo://nasa.gsfc.gcn/SWIFT#BAT_GRB_Pos_532871-729')

    def test_dumps(self):
        """
        Note, the processed output does not match the raw input -
        because I have added the UTF-8 encoding declaration.
        So we match the convenience routines against an
        etree.tostring processed version of the original.
        """
        swift_grb_v2_raw = objectify.parse(
            datapaths.swift_bat_grb_pos_v2).getroot()
        with open(datapaths.swift_bat_grb_pos_v2, 'rb') as f:
            swift_grb_v2_voeparsed = vp.load(f)
        raw = etree.tostring(swift_grb_v2_raw,
                             pretty_print=False,
                             xml_declaration=True,
                             encoding='UTF-8')
        processed = vp.dumps(swift_grb_v2_voeparsed)
        self.assertEqual(raw, processed)


class TestMinimalVOEvent(TestCase):
    def test_make_minimal_voevent(self):
        v1 = vp.Voevent(stream='voevent.soton.ac.uk/TEST',
                         stream_id='100',
                         role='test')
        self.assertTrue(vp.valid_as_v2_0(v1))
        v2 = vp.Voevent(stream='voevent.soton.ac.uk/TEST',
                         stream_id=100,
                         role='test')
        self.assertEqual(v1.attrib['ivorn'], v2.attrib['ivorn'])


class TestWho(TestCase):
    def setUp(self):
        self.v = vp.Voevent(stream='voevent.soton.ac.uk/TEST',
                             stream_id=100,
                             role='test')
        self.date = datetime.datetime.utcnow()


    def test_set_who_date(self):
        vp.set_who(self.v, self.date)
        self.assertTrue(vp.valid_as_v2_0(self.v))

    def test_set_who_minimal(self):
        vp.set_who(self.v, self.date, author_ivorn='voevent.soton.ac.uk/TEST')
        self.assertTrue(vp.valid_as_v2_0(self.v))

    def test_set_author(self):
        vp.set_author(self.v, title='4 Pi Sky Project',
                       shortName='4PiSky',
                       contactName='Tim Staley',
                       contactEmail='tim.staley@soton.ac.uk',
                       contactPhone='123456789',
                       contributor='Bob')
        self.assertTrue(vp.valid_as_v2_0(self.v))
        self.assertEqual(self.v.Who.Author.shortName, '4PiSky')


class TestWhat(TestCase):
    def shortDescription(self):
        return None

    def setUp(self):
        self.v = vp.Voevent(stream='voevent.soton.ac.uk/TEST',
                             stream_id='100',
                             role='test')

    def test_autoconvert_off(self):
        """Param values can only be strings..."""
        self.v.What.append(vp.Param(name='Dead Parrot', ac=False))
        self.v.What.append(
            vp.Param(name='The Answer', value=str(42), ac=False))
        self.assertTrue(vp.valid_as_v2_0(self.v))

        with self.assertRaises(TypeError):
            self.v.What.append(vp.Param(name='IntValue', value=42, ac=False))


    def test_autoconvert_on(self):
        """...but we provide some python smarts to alleviate this."""
        self.v.What.append(vp.Param(name='Dead Parrot'))
        self.v.What.append(vp.Param(name='The Answer', value=42))
        self.v.What.append(vp.Param(name='What is the time?',
                                     value=datetime.datetime.utcnow()))
        self.v.What.append(vp.Param(name='This is a lie',
                                     value=False))
        self.assertTrue(vp.valid_as_v2_0(self.v))


# print
#         print voe.prettystr(self.v.What)


class TestWhereWhen(TestCase):
    def setUp(self):
        self.v = vp.Voevent(stream='voevent.soton.ac.uk/TEST',
                             stream_id='100',
                             role='test')
        self.coords1 = vp.Position2D(ra=123.456, dec=45.678,
                                     err=0.1,
                                     units='deg', system='UTC-FK5-GEO'
        )

        self.coords2 = vp.Position2D(ra=355.456, dec=-57.678,
                                     err=0.1,
                                     units='deg', system='UTC-FK5-GEO')

    def test_set_wherewhen(self):
        vp.add_where_when(self.v, coords=self.coords1,
                           obs_time=datetime.datetime.utcnow(),
                           observatory_location=vp.definitions.observatory_location.geosurface)
        self.assertTrue(vp.valid_as_v2_0(self.v))
        self.assertEqual(self.coords1, vp.pull_astro_coords(self.v))

    def test_multiple_obs(self):
        self.assertEqual(self.v.WhereWhen.countchildren(), 0)
        vp.add_where_when(self.v, coords=self.coords1,
                           obs_time=datetime.datetime.utcnow(),
                           observatory_location=vp.definitions.observatory_location.geosurface)
        self.assertEqual(self.v.WhereWhen.countchildren(), 1)
        vp.add_where_when(self.v, coords=self.coords2,
                           obs_time=datetime.datetime.utcnow(),
                           observatory_location=vp.definitions.observatory_location.geosurface)
        self.assertEqual(self.v.WhereWhen.countchildren(), 2)
        self.assertTrue(vp.valid_as_v2_0(self.v))

        #How to reset to empty state?
        self.v.WhereWhen.ObsDataLocation=[]
        self.assertEqual(self.v.WhereWhen.countchildren(), 0)
        self.assertTrue(vp.valid_as_v2_0(self.v))
        # print vp.prettystr(self.v.WhereWhen)
        # print vp.dumps(self.v)

        vp.add_where_when(self.v, coords=self.coords2,
                           obs_time=datetime.datetime.utcnow(),
                           observatory_location=vp.definitions.observatory_location.geosurface)
        self.assertEqual(self.v.WhereWhen.countchildren(), 1)
        self.assertTrue(vp.valid_as_v2_0(self.v))


class TestHow(TestCase):
    def setUp(self):
        self.v = vp.Voevent(stream='voevent.soton.ac.uk/TEST',
                             stream_id='100',
                             role='test')

    def test_add_How(self):
        descriptions = ['One sentence.', 'Another.']
        vp.add_how(self.v, descriptions)
        self.assertEqual(len(self.v.How.Description), 2)
        self.assertEqual(descriptions,
                         [self.v.How.Description[0], self.v.How.Description[1]])
        refs = [vp.Reference('http://www.saltycrane.com/blog/2011/07/'
                              'example-parsing-xml-lxml-objectify/'),
                vp.Reference('http://github.com/timstaley/voevent-parse')]
        vp.add_how(self.v, references=refs)
        self.assertEqual(len(self.v.How.Reference), len(refs))
        self.assertEqual([r.attrib['uri'] for r in refs],
                         [r.attrib['uri'] for r in self.v.How.Reference])

        self.assertTrue(vp.valid_as_v2_0(self.v))


class TestWhy(TestCase):
    def setUp(self):
        self.v = vp.Voevent(stream='voevent.soton.ac.uk/TEST',
                             stream_id='100',
                             role='test')

    def test_add_why(self):
        inferences = [vp.Inference(probability=0.5, relation=None,
                                    name='Toin Coss', concept='Probability')]
        vp.add_why(self.v, importance=0.6,
                    expires=datetime.datetime(2013, 1, 1),
                    inferences=inferences)
        self.assertTrue(vp.valid_as_v2_0(self.v))
        self.assertEqual(self.v.Why.attrib['importance'], str(0.6))
        self.assertEqual(self.v.Why.Inference[0].attrib['probability'],
                         str(0.5))
        self.assertEqual(self.v.Why.Inference[0].Name, 'Toin Coss')


class TestCitations(TestCase):
    def setUp(self):
        self.v = vp.Voevent(stream='voevent.soton.ac.uk/TEST',
                             stream_id='100',
                             role='test')

    def test_followup_citation(self):
        ref = 'ivo://nasa.gsfc.gcn/SWIFT#BAT_GRB_Pos_532871-729'
        vp.add_citations(self.v,
                          vp.Citation(
                              'ivo://nasa.gsfc.gcn/SWIFT#BAT_GRB_Pos_532871-729',
                              cite_type=vp.definitions.cite_types.followup)
        )
        vp.assert_valid_as_v2_0(self.v)
        self.assertEqual(len(self.v.Citations.getchildren()), 1)
        # print
        # print vp.prettystr(self.v.Citations)
        vp.add_citations(self.v,
                          vp.Citation(
                              'ivo://nasa.gsfc.gcn/SWIFT#BAT_GRB_Pos_532871-730',
                              cite_type=vp.definitions.cite_types.followup)
        )
        self.assertTrue(vp.valid_as_v2_0(self.v))
        #         print voe.prettystr(self.v.Citations)
        self.assertEqual(len(self.v.Citations.getchildren()), 2)


class TestConvenienceRoutines(TestCase):
    def setUp(self):
        with open(datapaths.swift_bat_grb_pos_v2, 'rb') as f:
            self.swift_grb_v2_packet = vp.load(f)
        with open(datapaths.moa_lensing_event_path, 'rb') as f:
            self.moa_packet = vp.load(f)
        self.blank = vp.Voevent(stream='voevent.soton.ac.uk/TEST',
                                 stream_id='100',
                                 role='test')

    def test_pull_astro_coords(self):
        known_swift_grb_posn = vp.Position2D(ra=74.741200, dec=-9.313700,
                                              err=0.05,
                                              units='deg',
                                              system='UTC-FK5-GEO')
        p = vp.pull_astro_coords(self.swift_grb_v2_packet)
        self.assertEqual(p, known_swift_grb_posn)
        self.assertIsInstance(p.ra, float)

    def test_pull_params(self):
        params = vp.pull_params(self.swift_grb_v2_packet)

        #General example, check basic functionality
        self.assertEqual(params[None]['Packet_Type']['value'], '61')
        self.assertEqual(params['Misc_Flags']['Values_Out_of_Range']['value'],
                         'false')

        #Test empty What section
        params = vp.pull_params(self.blank)
        self.assertEqual(params, {})

        #Test known (generated) example
        single_par = copy(self.blank)
        single_par.What.append(vp.Param(name="test_param", value=123))
        params = vp.pull_params(single_par)
        self.assertEqual(len(params), 1)
        self.assertEqual(list(params[None].keys()), ['test_param'])

        #Test case where a What Group is present, but empty:
        params = vp.pull_params(self.moa_packet)
        self.assertEqual(params['Misc_Flags'], {})


    def test_pull_isotime(self):
        isotime = vp.pull_isotime(self.swift_grb_v2_packet)
        self.assertIsInstance(isotime, datetime.datetime)

        null_result = vp.pull_isotime(self.blank)
        self.assertIsNone(null_result)




import os
from voeventparse.tests.resources import __path__ as data_dir

data_dir = data_dir[0]
swift_bat_grb_pos_v2 = os.path.join(data_dir,
                                    'SWIFT_bat_position_v2.0_example.xml')
swift_xrt_pos_v1 = os.path.join(data_dir,
                                    'SWIFT_xrt_position_v1.1_example.xml')
no_namespace_test_packet = os.path.join(data_dir, 'no_namespace_packet.xml')

moa_lensing_event_path = os.path.join(data_dir,
                                'MOA_Lensing_Event_2015-07-10T14:50:54.00.xml')



"""Check the syntax in example files is valid"""

import unittest
from unittest import TestCase
import tempfile
import os
import sys


class TestExamples(TestCase):
    def setUp(self):
        # Run in a tempdir, in case the examples dump any output
        self.orig_dir = os.getcwd()
        self.tempdir = tempfile.mkdtemp()
        os.chdir(self.tempdir)

        repo_topdir = os.path.dirname(
            os.path.dirname(os.path.dirname(__file__)))
        self.examples_dir = os.path.join(repo_topdir, 'examples')
        sys.path.insert(0, self.examples_dir)

    def tearDown(self):
        os.chdir(self.orig_dir)
        sys.path.pop(0)

    def test_basic_usage(self):
        import basic_usage

    def test_new_voevent(self):
        import author_new_voevent








"""This module simply serves to store the XML schema, a 'skeleton' VOEvent
xml document for creation of new instances, and various other minor definitions.

These values may be used in place of literal strings, to allow autocompletion
and document the fact that they are 'standardized' values.
"""
from __future__ import unicode_literals
from six import b

# ############################
# Some useful string defs namespaced via a class container:

class roles(object):
    observation = 'observation'
    prediction = 'prediction'
    utility = 'utility'
    test = 'test'


class sky_coord_system(object):
    """
    Common coordinate system identifiers. See also :class:`.Position2D`.

    This is not a simple combinatorial mix of all components, it's a
    reproduction of the enumeration in the XML schema. Note some entries in the
    schema enumeration are repeated, which leads me to think it may be flawed,
    but that's an investigation for another day.

    """
    gps_fk5_topo = 'GPS-FK5-TOPO'
    gps_icrs_geo = 'GPS-ICRS-GEO'
    gps_icrs_topo = 'GPS-ICRS-TOPO'

    tdb_fk5_bary = 'TDB-FK5-BARY'
    tdb_icrs_bary = 'TDB-ICRS-BARY'

    tt_fk5_geo = 'TT-FK5-GEO'
    tt_fk5_topo = 'TT-FK5-TOPO'
    tt_icrs_geo = 'TT-ICRS-GEO'
    tt_icrs_topo = 'TT-ICRS-TOPO'

    utc_fk5_geo = 'UTC-FK5-GEO'
    utc_fk5_topo = 'UTC-FK5-TOPO'
    utc_icrs_geo = 'UTC-ICRS-GEO'
    utc_icrs_topo = 'UTC-ICRS-TOPO'




class observatory_location(object):
    """Common generic values for the WhereWhen.ObservatoryLocation attribute."""
    geosurface = 'GEOSURFACE'
    geolunar = 'GEOLUN'


class coord_units(object):
    """Handy tags listing the unit names used by voeventparse."""
    degrees = 'degrees'


class cite_types(object):
    """Possible types of :func:`.Citation`"""
    followup = 'followup'
    supersedes = 'supersedes'
    retraction = 'retraction'

####################################
#: Skeleton XML for instantiating a new VOEvent tree
v2_0_skeleton_str = b"""<?xml version="1.0" ?>
<voe:VOEvent xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
    xmlns:voe="http://www.ivoa.net/xml/VOEvent/v2.0"
    xsi:schemaLocation="http://www.ivoa.net/xml/VOEvent/v2.0 http://www.ivoa.net/xml/VOEvent/VOEvent-v2.0.xsd"
    version="2.0"
    role="test"
    ivorn="ivo://undefined">
</voe:VOEvent>
"""

####################################
#: VOEvent 2.0 schema for validation
#Schema downloaded from:
#http://www.ivoa.net/xml/VOEvent/VOEvent-v2.0.xsd
v2_0_schema_str = b"""<?xml version="1.0"?>
<xs:schema xmlns="http://www.ivoa.net/xml/VOEvent/v2.0" xmlns:xs="http://www.w3.org/2001/XMLSchema"
  targetNamespace="http://www.ivoa.net/xml/VOEvent/v2.0" elementFormDefault="unqualified">

  <xs:element name="VOEvent">
    <xs:annotation>
      <xs:documentation> VOEvent is the root element for describing observations of immediate
        astronomical events. For more information, see
        http://www.ivoa.net/twiki/bin/view/IVOA/IvoaVOEvent. The event consists of at most one of
        each of: Who, What, WhereWhen, How, Why, Citations, Description, and
        Reference.</xs:documentation>
    </xs:annotation>
    <xs:complexType>
      <!-- Can have zero or one of each of these, in any order -->
      <xs:all>
        <xs:element name="Who"         type="Who"       minOccurs="0" maxOccurs="1"/>
        <xs:element name="What"        type="What"      minOccurs="0" maxOccurs="1"/>
        <xs:element name="WhereWhen"   type="WhereWhen" minOccurs="0" maxOccurs="1"/>
        <xs:element name="How"         type="How"       minOccurs="0" maxOccurs="1"/>
        <xs:element name="Why"         type="Why"       minOccurs="0" maxOccurs="1"/>
        <xs:element name="Citations"   type="Citations" minOccurs="0" maxOccurs="1"/>
        <xs:element name="Description" type="xs:string" minOccurs="0" maxOccurs="1"/>
        <xs:element name="Reference"   type="Reference" minOccurs="0" maxOccurs="1"/>
      </xs:all>
      <xs:attribute name="version"     type="xs:token"   fixed="2.0" use="required"/>
      <xs:attribute name="ivorn"       type="xs:anyURI"  use="required"/>
      <xs:attribute name="role"        type="roleValues" default="observation"/>
    </xs:complexType>
  </xs:element>
  <xs:simpleType name="roleValues">
    <xs:restriction base="xs:string">
      <xs:enumeration value="observation"/>
      <xs:enumeration value="prediction"/>
      <xs:enumeration value="utility"/>
      <xs:enumeration value="test"/>
    </xs:restriction>
  </xs:simpleType>


  <xs:complexType name="Who">
    <xs:annotation>
      <xs:documentation> Who: Curation Metadata </xs:documentation>
    </xs:annotation>
    <xs:all>
      <!-- Can have zero or one of each of these. Schema is loose because
        should have AuthorIVORN *or* Author -->
      <xs:element name="AuthorIVORN" type="xs:anyURI" minOccurs="0"/>
      <xs:element name="Date" type="xs:dateTime" minOccurs="0"/>
      <xs:element name="Description" type="xs:string" minOccurs="0"/>
      <xs:element name="Reference" type="Reference" minOccurs="0"/>
      <xs:element name="Author" minOccurs="0">
        <xs:annotation>
          <xs:documentation> Author information follows the IVOA curation information schema: the
            organization responsible for the packet can have a title, short name or acronym, and a
            logo. A contact person has a name, email, and phone number. Other contributors can also
            be noted. </xs:documentation>
        </xs:annotation>
        <xs:complexType>
          <xs:choice maxOccurs="unbounded">
            <xs:element name="title" type="xs:string"/>
            <xs:element name="shortName" type="xs:string"/>
            <xs:element name="logoURL" type="xs:anyURI"/>
            <xs:element name="contactName" type="xs:string"/>
            <xs:element name="contactEmail" type="xs:string"/>
            <xs:element name="contactPhone" type="xs:string"/>
            <xs:element name="contributor" type="xs:string"/>
          </xs:choice>
        </xs:complexType>
      </xs:element>
    </xs:all>
  </xs:complexType>


  <xs:complexType name="What">
    <xs:annotation>
      <xs:documentation> What: Event Characterization. This is the part of the data model that is
        chosen by the Authoer of the event rather than the IVOA. There can be Params, that may be in
        Groups, and Tables, and simpleTimeSeries. There can also be Description and Reference as
        with most VOEvent elements. </xs:documentation>
    </xs:annotation>
    <xs:choice maxOccurs="unbounded">
      <!-- can have any number of any of these in any order  -->
      <xs:element name="Param" type="Param" minOccurs="0" maxOccurs="unbounded"/>
      <xs:element name="Group" type="Group" minOccurs="0" maxOccurs="unbounded"/>
      <xs:element name="Table" type="Table" minOccurs="0" maxOccurs="unbounded"/>
      <!--<xs:element ref="sts:SimpleTimeseries"          minOccurs="0" maxOccurs="unbounded"/>-->
      <xs:element name="Description" type="xs:string" minOccurs="0" maxOccurs="unbounded"/>
      <xs:element name="Reference" type="Reference" minOccurs="0" maxOccurs="unbounded"/>
    </xs:choice>
  </xs:complexType>


  <xs:complexType name="Param">
    <xs:annotation>
      <xs:documentation> What/Param definition. A Param has name, value, ucd, unit, dataType; and
        may have Description and Reference.</xs:documentation>
    </xs:annotation>
    <xs:choice maxOccurs="unbounded">
      <xs:element name="Description" type="xs:string" minOccurs="0"/>
      <xs:element name="Reference" type="Reference" minOccurs="0" maxOccurs="unbounded"/>
      <xs:element name="Value" type="xs:string" minOccurs="0" maxOccurs="1"/>
    </xs:choice>
    <xs:attribute name="name" type="xs:string"/>
    <xs:attribute name="ucd" type="xs:string"/>
    <xs:attribute name="value" type="xs:string"/>
    <xs:attribute name="unit" type="xs:string"/>
    <xs:attribute name="dataType" type="dataType" default="string"/>
    <xs:attribute name="utype" type="xs:string"/>
  </xs:complexType>
  <xs:simpleType name="dataType">
    <xs:restriction base="xs:string">
      <xs:enumeration value="string"/>
      <xs:enumeration value="float"/>
      <xs:enumeration value="int"/>
    </xs:restriction>
  </xs:simpleType>

  <xs:complexType name="Group">
    <xs:annotation>
      <xs:documentation> What/Group definition: A group is a collection of Params, with name and
        type attributes.</xs:documentation>
    </xs:annotation>
    <xs:choice maxOccurs="unbounded">
      <xs:element name="Param" type="Param" maxOccurs="unbounded"/>
      <xs:element name="Description" type="xs:string" minOccurs="0"/>
      <xs:element name="Reference" type="Reference" minOccurs="0"/>
    </xs:choice>
    <xs:attribute name="name" type="xs:string"/>
    <xs:attribute name="type" type="xs:string"/>
  </xs:complexType>


  <xs:complexType name="Table">
    <xs:annotation>
      <xs:documentation> What/Table definition. This small Table has Fields for the column
        definitions, and Data to hold the table data, with TR for row and TD for value of a table
        cell.</xs:documentation>
    </xs:annotation>
    <xs:choice maxOccurs="unbounded">
      <xs:element name="Description" type="xs:string" minOccurs="0"/>
      <xs:element name="Reference" type="Reference" minOccurs="0"/>
      <xs:element name="Param" type="Param" minOccurs="0" maxOccurs="unbounded"/>
      <xs:element name="Field" type="Field" minOccurs="0" maxOccurs="unbounded"/>
      <xs:element name="Data" type="Data" minOccurs="1" maxOccurs="1"/>
    </xs:choice>
    <xs:attribute name="name" type="xs:string"/>
    <xs:attribute name="type" type="xs:string"/>
  </xs:complexType>
  <xs:complexType name="Field">
    <xs:choice maxOccurs="unbounded">
      <xs:element name="Description" type="xs:string" minOccurs="0"/>
      <xs:element name="Reference" type="Reference" minOccurs="0"/>
    </xs:choice>
    <xs:attribute name="name" type="xs:string"/>
    <xs:attribute name="ucd" type="xs:string"/>
    <xs:attribute name="unit" type="xs:string"/>
    <xs:attribute name="dataType" type="dataType" default="string"/>
    <xs:attribute name="utype" type="xs:string"/>
  </xs:complexType>
  <xs:complexType name="Data">
    <xs:choice maxOccurs="unbounded">
      <xs:element name="TR" type="TR"/>
    </xs:choice>
  </xs:complexType>
  <xs:complexType name="TR">
    <xs:choice maxOccurs="unbounded">
      <xs:element name="TD" type="xs:string"/>
    </xs:choice>
  </xs:complexType>


  <xs:complexType name="WhereWhen">
    <xs:annotation>
      <xs:documentation> WhereWhen: Space-Time Coordinates. Lots and lots of elements here, but the
        import is that each event has these: observatory, sky_coord_system, time, timeError, longitude,
        latitude, posError.</xs:documentation>
    </xs:annotation>
    <xs:choice maxOccurs="unbounded">
      <xs:element name="ObsDataLocation" type="ObsDataLocation" minOccurs="1" maxOccurs="1"/>
      <xs:element name="Description" type="xs:string" minOccurs="0"/>
      <xs:element name="Reference" type="Reference" minOccurs="0"/>
    </xs:choice>
    <xs:attribute name="id" type="xs:ID" use="optional"/>
  </xs:complexType>

  <xs:complexType name="ObsDataLocation">
    <xs:annotation>
      <xs:documentation> Part of WhereWhen</xs:documentation>
    </xs:annotation>
    <xs:all>
      <xs:element name="ObservatoryLocation" minOccurs="1" maxOccurs="1" type="ObservatoryLocation"/>
      <xs:element name="ObservationLocation" minOccurs="1" maxOccurs="1" type="ObservationLocation"/>
    </xs:all>
  </xs:complexType>

  <xs:complexType name="ObservationLocation">
    <xs:annotation>
      <xs:documentation> Part of WhereWhen</xs:documentation>
    </xs:annotation>
    <xs:all>
      <xs:element name="AstroCoordSystem" type="AstroCoordSystem" minOccurs="1" maxOccurs="1"/>
      <xs:element name="AstroCoords" type="AstroCoords" minOccurs="1" maxOccurs="1"/>
    </xs:all>
  </xs:complexType>

  <xs:complexType name="AstroCoordSystem">
    <xs:annotation>
      <xs:documentation> Part of WhereWhen</xs:documentation>
    </xs:annotation>
    <!-- The empty sequence closes this to content -->
    <xs:sequence />
    <xs:attribute name="id" type="idValues"/>
  </xs:complexType>
  <xs:simpleType name="idValues">
    <xs:restriction base="xs:string">
      <xs:enumeration value="TT-ICRS-TOPO"/>
      <xs:enumeration value="UTC-ICRS-TOPO"/>
      <xs:enumeration value="TT-FK5-TOPO"/>
      <xs:enumeration value="UTC-FK5-TOPO"/>
      <xs:enumeration value="GPS-ICRS-TOPO"/>
      <xs:enumeration value="GPS-ICRS-TOPO"/>
      <xs:enumeration value="GPS-FK5-TOPO"/>
      <xs:enumeration value="GPS-FK5-TOPO"/>
      <xs:enumeration value="TT-ICRS-GEO"/>
      <xs:enumeration value="UTC-ICRS-GEO"/>
      <xs:enumeration value="TT-FK5-GEO"/>
      <xs:enumeration value="UTC-FK5-GEO"/>
      <xs:enumeration value="GPS-ICRS-GEO"/>
      <xs:enumeration value="GPS-ICRS-GEO"/>
      <xs:enumeration value="TDB-ICRS-BARY"/>
      <xs:enumeration value="TDB-FK5-BARY"/>
      <!-- this one for ObservatoryLocation -->
      <xs:enumeration value="UTC-GEOD-TOPO"/>
    </xs:restriction>
  </xs:simpleType>

  <xs:complexType name="AstroCoords">
    <xs:annotation>
      <xs:documentation> Part of WhereWhen</xs:documentation>
    </xs:annotation>
    <xs:all>
      <xs:element name="Time" type="Time" minOccurs="0"/>
      <xs:element name="Position2D" type="Position2D" minOccurs="0"/>
      <xs:element name="Position3D" type="Position3D" minOccurs="0"/>
    </xs:all>
    <xs:attribute name="coord_system_id" type="idValues"/>
  </xs:complexType>

  <xs:complexType name="Time">
    <xs:annotation>
      <xs:documentation> Part of WhereWhen</xs:documentation>
    </xs:annotation>
    <xs:choice maxOccurs="unbounded">
      <xs:element name="TimeInstant" type="TimeInstant"/>
      <xs:element name="Error" type="xs:float" minOccurs="0"/>
    </xs:choice>
    <xs:attribute name="unit" type="xs:string"/>
  </xs:complexType>

  <xs:complexType name="TimeInstant">
    <xs:annotation>
      <xs:documentation> Part of WhereWhen</xs:documentation>
    </xs:annotation>
    <xs:choice maxOccurs="unbounded">
      <xs:element name="ISOTime" type="xs:string" minOccurs="0" maxOccurs="1"/>
      <xs:element name="TimeOffset" type="xs:float" minOccurs="0" maxOccurs="1"/>
      <xs:element name="TimeScale" type="xs:string" minOccurs="0" maxOccurs="1"/>
    </xs:choice>
  </xs:complexType>

  <xs:complexType name="Position2D">
    <xs:annotation>
      <xs:documentation> Part of WhereWhen</xs:documentation>
    </xs:annotation>
    <xs:all>
      <xs:element name="Name1" type="xs:string" minOccurs="0"/>
      <xs:element name="Name2" type="xs:string" minOccurs="0"/>
      <xs:element name="Value2" type="Value2"/>
      <xs:element name="Error2Radius" type="xs:float"/>
    </xs:all>
    <xs:attribute name="unit" type="xs:string"/>
  </xs:complexType>

  <xs:complexType name="Position3D">
    <xs:annotation>
      <xs:documentation> Part of WhereWhen</xs:documentation>
    </xs:annotation>
    <xs:all>
      <xs:element name="Name1" type="xs:string" minOccurs="0"/>
      <xs:element name="Name2" type="xs:string" minOccurs="0"/>
      <xs:element name="Name3" type="xs:string" minOccurs="0"/>
      <xs:element name="Value3" type="Value3"/>
    </xs:all>
    <xs:attribute name="unit" type="xs:string"/>
  </xs:complexType>

  <xs:complexType name="Value2">
    <xs:annotation>
      <xs:documentation> Part of WhereWhen</xs:documentation>
    </xs:annotation>
    <xs:all>
      <xs:element name="C1" type="xs:float"/>
      <xs:element name="C2" type="xs:float"/>
    </xs:all>
  </xs:complexType>

  <xs:complexType name="Value3">
    <xs:annotation>
      <xs:documentation> Part of WhereWhen</xs:documentation>
    </xs:annotation>
    <xs:all>
      <xs:element name="C1" type="xs:float"/>
      <xs:element name="C2" type="xs:float"/>
      <xs:element name="C3" type="xs:float"/>
    </xs:all>
  </xs:complexType>

  <xs:complexType name="ObservatoryLocation">
    <xs:annotation>
      <xs:documentation> Part of WhereWhen</xs:documentation>
    </xs:annotation>
    <xs:all>
      <xs:element name="AstroCoordSystem" type="AstroCoordSystem" minOccurs="0" maxOccurs="1"/>
      <xs:element name="AstroCoords" type="AstroCoords" minOccurs="0" maxOccurs="1"/>
    </xs:all>
    <xs:attribute name="id" type="xs:string"/>
  </xs:complexType>

  <xs:complexType name="How">
    <xs:annotation>
      <xs:documentation> How: Instrument Configuration. Built with some Description and Reference
        elements. </xs:documentation>
    </xs:annotation>
    <xs:choice maxOccurs="unbounded">
      <xs:element name="Description" type="xs:string"/>
      <xs:element name="Reference" type="Reference"/>
    </xs:choice>
  </xs:complexType>

  <xs:complexType name="Why">
    <xs:annotation>
      <xs:documentation> Why: Initial Scientific Assessment. Can make simple Concept/Name/Desc/Ref
        for the inference or use multiple Inference containers for more semantic sophistication.
      </xs:documentation>
    </xs:annotation>
    <xs:choice maxOccurs="unbounded">
      <xs:element name="Name" type="xs:string"/>
      <xs:element name="Concept" type="xs:string"/>
      <xs:element name="Inference" type="Inference"/>
      <xs:element name="Description" type="xs:string"/>
      <xs:element name="Reference" type="Reference"/>
    </xs:choice>
    <xs:attribute name="importance" type="xs:float"/>
    <xs:attribute name="expires" type="xs:dateTime"/>
  </xs:complexType>

  <xs:complexType name="Inference">
    <xs:annotation>
      <xs:documentation> Why/Inference: A container for a more nuanced expression, including
        relationships and probability. </xs:documentation>
    </xs:annotation>
    <xs:choice maxOccurs="unbounded">
      <xs:element name="Name" type="xs:string"/>
      <xs:element name="Concept" type="xs:string"/>
      <xs:element name="Description" type="xs:string"/>
      <xs:element name="Reference" type="Reference"/>
    </xs:choice>
    <xs:attribute name="probability" type="smallFloat"/>
    <xs:attribute name="relation" type="xs:string"/>
  </xs:complexType>
  <xs:simpleType name="smallFloat">
    <xs:restriction base="xs:float">
      <xs:minInclusive value="0.0"/>
      <xs:maxInclusive value="1.0"/>
    </xs:restriction>
  </xs:simpleType>


  <xs:complexType name="Citations">
    <xs:annotation>
      <xs:documentation> Citations: Follow-up Observations. This section is a sequence of EventIVORN
        elements, each of which has the IVORN of a cited event. </xs:documentation>
    </xs:annotation>
    <xs:sequence>
      <xs:element name="EventIVORN" type="EventIVORN" maxOccurs="unbounded"/>
      <xs:element name="Description" type="xs:string" minOccurs="0"/>
    </xs:sequence>
  </xs:complexType>

  <xs:complexType name="EventIVORN">
    <xs:annotation>
      <xs:documentation> Citations/EventIVORN. The value is the IVORN of the cited event, the 'cite'
        attribute is the nature of that relationship, choosing from 'followup', 'supersedes', or
        'retraction'.</xs:documentation>
    </xs:annotation>
    <xs:simpleContent>
      <xs:extension base="xs:string">
        <xs:attribute name="cite" type="citeValues"/>
      </xs:extension>
    </xs:simpleContent>
  </xs:complexType>

  <xs:simpleType name="citeValues">
    <xs:restriction base="xs:string">
      <xs:enumeration value="followup"/>
      <xs:enumeration value="supersedes"/>
      <xs:enumeration value="retraction"/>
    </xs:restriction>
  </xs:simpleType>

  <xs:complexType name="Reference">
    <xs:annotation>
      <xs:documentation> Reference: External Content. The payload is the uri, and the 'type'
        describes the nature of the data under that uri. The Reference can also be named.
      </xs:documentation>
    </xs:annotation>
    <xs:sequence/>
    <xs:attribute name="uri" type="xs:anyURI" use="required"/>
    <xs:attribute name="type" type="xs:string"/>
    <xs:attribute name="mimetype" type="xs:string"/>
    <xs:attribute name="meaning" type="xs:anyURI"/>
  </xs:complexType>

</xs:schema>
"""



"""Convenience routines for common actions on VOEvent objects"""

from __future__ import absolute_import
import iso8601
import lxml
from voeventparse.misc import (Param, Group, Reference, Inference, Position2D,
                               Citation)



def pull_astro_coords(voevent, index=0):
    """Extracts the `AstroCoords` from a given `WhereWhen.ObsDataLocation`.

    Note that a packet may include multiple 'ObsDataLocation' entries
    under the 'WhereWhen' section, for example giving locations of an object
    moving over time. Most packets will have only one, however, so the
    default is to just return co-ords extracted from the first.

    Args:
        voevent (:class:`voeventparse.voevent.Voevent`): Root node of the
            VOEvent etree.
        index (int): Index of the ObsDataLocation to extract AstroCoords from.

    Returns:
        Position (:py:class:`.Position2D`): The sky position defined in the
        ObsDataLocation.
    """
    od = voevent.WhereWhen.ObsDataLocation[index]
    ac = od.ObservationLocation.AstroCoords
    ac_sys = voevent.WhereWhen.ObsDataLocation.ObservationLocation.AstroCoordSystem
    sys = ac_sys.attrib['id']


    if hasattr(ac.Position2D, "Name1"):
        assert ac.Position2D.Name1 == 'RA' and ac.Position2D.Name2 == 'Dec'
    posn = Position2D(ra=float(ac.Position2D.Value2.C1),
                      dec=float(ac.Position2D.Value2.C2),
                      err=float(ac.Position2D.Error2Radius),
                      units=ac.Position2D.attrib['unit'],
                      system=sys)
    return posn


def pull_isotime(voevent, index=0):
    """Extracts the event time from a given `WhereWhen.ObsDataLocation`.

    Accesses a `WhereWhere.ObsDataLocation.ObservationLocation`
    element and returns the AstroCoords.Time.TimeInstant.ISOTime element,
    converted to a (UTC-timezoned) datetime.

    Note that a packet may include multiple 'ObsDataLocation' entries
    under the 'WhereWhen' section, for example giving locations of an object
    moving over time. Most packets will have only one, however, so the
    default is to access the first.

    .. warning::

        This function currently only works with UTC time-system coords.
        Future updates may implement conversion from other systems (TT, GPS)
        using astropy functionality.

    Args:
        voevent (:class:`voeventparse.voevent.Voevent`): Root node of the VOevent
            etree.
        index (int): Index of the ObsDataLocation to extract an ISOtime from.

    Returns:
        isotime (:class:`datetime.datetime`): Datetime object as parsed by
        `iso8601`_ (with UTC timezone).

    .. _iso8601: https://pypi.python.org/pypi/iso8601/

    """
    try:
        od = voevent.WhereWhen.ObsDataLocation[index]
        ol = od.ObservationLocation
        coord_sys = ol.AstroCoords.attrib['coord_system_id']
        if coord_sys.split('-')[0] != 'UTC':
            raise NotImplementedError(
                'Loading from time-systems other than UTC not yet implemented'
            )
        isotime_str = str(ol.AstroCoords.Time.TimeInstant.ISOTime)
        return iso8601.parse_date(isotime_str)
    except AttributeError:
        return None


def pull_params(voevent):
    """Attempts to load the `What` section of a voevent as a nested dictionary.

    Args:
        voevent (:class:`voeventparse.voevent.Voevent`): Root node of the VOevent etree.
    Returns:
        Nested dict (dict): Mapping of ``Group->Param->Attribs``.
        Access like so::

            foo_param_val = what_dict['GroupName']['ParamName']['value']

        .. note::

          Parameters without a group are indexed under the key 'None' - otherwise,
          we might get name-clashes between `params` and `groups` (unlikely but
          possible) so for ungrouped Params you'll need something like::

            what_dict[None]['ParamName']['value']

    """
    result = {}
    w = voevent.What
    if w.countchildren() == 0:
        return result
    toplevel_params = {}
    result[None] = toplevel_params
    if hasattr(voevent.What, 'Param'):
        for p in voevent.What.Param:
            toplevel_params[p.attrib['name']] = p.attrib
    if hasattr(voevent.What, 'Group'):
        for g in voevent.What.Group:
            g_params = {}
            result[g.attrib['name']] = g_params
            if hasattr(g, 'Param'):
                for p in g.Param:
                    g_params[p.attrib['name']] = p.attrib
    return result


def prettystr(subtree):
    """Print an element tree with nice indentation.

    Prettyprinting a whole VOEvent often doesn't seem to work, probably for
    issues relating to whitespace cf.
    http://lxml.de/FAQ.html#why-doesn-t-the-pretty-print-option-reformat-my-xml-output
    This function is a quick workaround for prettyprinting a subsection
    of a VOEvent, for easier desk-checking.

    Args:
        subtree(lxml.etree): A node in the VOEvent element tree.
    Returns:
        string: Prettyprinted string representation of the raw XML.
    """
    lxml.objectify.deannotate(subtree)
    lxml.etree.cleanup_namespaces(subtree)
    return lxml.etree.tostring(subtree, pretty_print=True)





"""
A package for concise manipulation of VOEvent XML packets.
"""

from __future__ import absolute_import
import os
from pkg_resources import get_distribution, DistributionNotFound

from voeventparse.voevent import (
    Voevent,
    voevent_v2_0_schema,
    load, loads, dump, dumps,
    valid_as_v2_0, assert_valid_as_v2_0,
    set_who, set_author, add_where_when,
    add_how, add_why, add_citations
)
import voeventparse.definitions as definitions
from voeventparse.misc import (
    Param, Group,
    Reference, Inference,
    Position2D,
    Citation)
from voeventparse.convenience import (
    pull_astro_coords, pull_params, pull_isotime,
    prettystr)



###########################################################
# Versioning; see also
# http://stackoverflow.com/questions/17583443
###########################################################
try:
    _dist = get_distribution('voevent-parse')
    #The version number according to Pip:
    _nominal_version = _dist.version
    if not __file__.startswith(os.path.join(_dist.location, 'voeventparse')):
        # not installed, but there is another version that *is*
        raise DistributionNotFound
except DistributionNotFound:
    #The actual copy in use if a custom PYTHONPATH or local dir import is used
    __version__ = 'Local import @ ' + os.path.dirname(os.path.abspath(__file__))
else:
    __version__ = _nominal_version


